#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
### This function inspired by Miguel A. Rodriguez                            ###
###  https://fenicsproject.org/qa/12933/                                     ###
###            making-vtk-python-object-from-solution-object-the-same-script ###
###                                                                          ###
################################################################################

import dolfin
import numpy
import vtk

import dolfin_mech as dmech

################################################################################

def mesh2ugrid(
        mesh,
        verbose=0):

    if (verbose): print("mesh2ugrid")

    n_dim = mesh.geometry().dim()
    assert (n_dim in (2,3))
    if (verbose): print("n_dim = "+str(n_dim))

    n_verts = mesh.num_vertices()
    if (verbose): print("n_verts = "+str(n_verts))
    n_cells = mesh.num_cells()
    if (verbose): print("n_cells = "+str(n_cells))

    # Create function space
    fe = dolfin.FiniteElement(
        family="CG",
        cell=mesh.ufl_cell(),
        degree=1,
        quad_scheme='default')
    fs = dolfin.FunctionSpace(
        mesh,
        fe)

    # Store nodes coordinates as numpy array
    n_nodes = fs.dim()
    assert (n_nodes == n_verts),\
        "n_nodes ("+str(n_nodes)+") ≠ n_verts ("+str(n_verts)+"). Aborting."
    if (verbose): print("n_nodes = "+str(n_nodes))
    global np_coordinates # MG20190621: if it disappears the vtk objects is broken
    np_coordinates = fs.tabulate_dof_coordinates().reshape([n_nodes, n_dim])
    if (verbose): print("np_coordinates = "+str(np_coordinates))

    if (n_dim == 2):
        np_coordinates = numpy.hstack((np_coordinates, numpy.zeros([n_nodes, 1])))
        if (verbose): print("np_coordinates = "+str(np_coordinates))

    # Convert nodes coordinates to VTK
    vtk_coordinates = vtk.util.numpy_support.numpy_to_vtk(
        num_array=np_coordinates,
        deep=1)
    vtk_points = vtk.vtkPoints()
    vtk_points.SetData(vtk_coordinates)
    if (verbose): print("n_points = "+str(vtk_points.GetNumberOfPoints()))

    # Store connectivity as numpy array
    if   (n_dim == 2):
        n_nodes_per_cell = 3
    elif (n_dim == 3):
        n_nodes_per_cell = 4
    if (verbose): print("n_nodes_per_cell = "+str(n_nodes_per_cell))
    global np_connectivity # MG20190621: if it disappears the vtk objects is broken
    np_connectivity = numpy.empty(
        [n_cells, 1+n_nodes_per_cell],
        dtype=int)
    for i in range(n_cells):
        np_connectivity[i, 0] = n_nodes_per_cell
        np_connectivity[i,1:] = fs.dofmap().cell_dofs(i)
    # if (verbose): print("np_connectivity = "+str(np_connectivity))

    # Add left column specifying number of nodes per cell and flatten array
    np_connectivity = np_connectivity.flatten()
    # if (verbose): print("np_connectivity = "+str(np_connectivity))

    # Convert connectivity to VTK
    vtk_connectivity = vtk.util.numpy_support.numpy_to_vtkIdTypeArray(np_connectivity)

    # Create cell array
    vtk_cells = vtk.vtkCellArray()
    vtk_cells.SetCells(n_cells, vtk_connectivity)
    if (verbose): print("n_cells = "+str(vtk_cells.GetNumberOfCells()))

    # Create unstructured grid and set points and connectivity
    if   (n_dim == 2):
        vtk_cell_type = vtk.VTK_TRIANGLE
    elif (n_dim == 3):
        vtk_cell_type = vtk.VTK_TETRA
    ugrid = vtk.vtkUnstructuredGrid()
    ugrid.SetPoints(vtk_points)
    ugrid.SetCells(vtk_cell_type, vtk_cells)

    return ugrid

################################################################################

def add_function_to_ugrid(
        function,
        ugrid,
        force_3d_field=1,
        verbose=0):

    if (verbose): print("add_function_to_ugrid")
    if (verbose): print(ugrid.GetPoints())

    # Convert function values and add as scalar data
    n_dofs = function.function_space().dim()
    if (verbose): print("n_dofs = "+str(n_dofs))
    n_dim = function.ufl_element().value_size()
    if (verbose): print("n_dim = "+str(n_dim))
    assert (n_dofs//n_dim == ugrid.GetNumberOfPoints()),\
        "Only CG1 functions can be converted to VTK. Aborting."
    global np_array # MG20190621: if it disappears the vtk object is broken
    np_array = function.vector().get_local()
    if (verbose): print("np_array = "+str(np_array))
    np_array = np_array.reshape([n_dofs//n_dim, n_dim])
    if (verbose): print("np_array = "+str(np_array))
    if (force_3d_field) and (n_dim == 2):
        np_array = numpy.hstack((np_array, numpy.zeros([n_dofs//n_dim, 1])))
        if (verbose): print("np_array = "+str(np_array))
    vtk_array = vtk.util.numpy_support.numpy_to_vtk(
        num_array=np_array,
        deep=1)
    vtk_array.SetName(function.name())

    if (verbose): print(ugrid.GetPoints())
    ugrid.GetPointData().AddArray(vtk_array)
    if (verbose): print(ugrid.GetPoints())

################################################################################

def add_functions_to_ugrid(
        functions,
        ugrid):

    for function in functions:
        dmech.add_function_to_ugrid(
            function=function,
            ugrid=ugrid)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Colin Laville, 2021-2022                                             ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

def get_ExprMeshFunction_cpp_pybind():

    cpp_code = """
#include <pybind11/pybind11.h>
#include <pybind11/eigen.h>

#include <dolfin/function/Expression.h>
#include <dolfin/mesh/MeshFunction.h>

class MeshExpr : public dolfin::Expression
{
public:
    // The data stored in mesh functions
    std::shared_ptr<dolfin::MeshFunction<double>> mf;

    // Create scalar expression
    MeshExpr() : dolfin::Expression() {}

    // Function for evaluating expression on each cell
    void eval(
        Eigen::Ref<Eigen::VectorXd> values,
        Eigen::Ref<const Eigen::VectorXd> x,
        const ufc::cell& cell) const override
    {
        const uint cell_index = cell.index;
        values[0] = (*mf)[cell_index];
    }
};

PYBIND11_MODULE(SIGNATURE, m)
{
pybind11::class_<MeshExpr, std::shared_ptr<MeshExpr>, dolfin::Expression>
(m, "MeshExpr")
.def(pybind11::init<>())
.def_readwrite("mf", &MeshExpr::mf);
}
"""

    return cpp_code



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import math
import numpy

import dolfin_mech as dmech

################################################################################

def run_Disc_Hyperelasticity(
        incomp=0,
        mesh_params={},
        mat_params={},
        step_params={},
        load_params={},
        res_basename="run_Disc_Hyperelasticity",
        write_vtus_with_preserved_connectivity=False,
        verbose=0):

    ################################################################### Mesh ###

    X0 = mesh_params.get("X0", 0.5)
    Y0 = mesh_params.get("Y0", 0.5)
    R  = mesh_params.get("R" , 0.3)

    mesh, boundaries_mf, S_id, points_mf, x1_sd, x2_sd, x3_sd, x4_sd = dmech.run_Disc_Mesh(
        params=mesh_params)

    ################################################################ Problem ###

    if (incomp):
        displacement_degree = 2 # MG20211219: Incompressibility requires displacement_degree >= 2 ?!
    else:
        displacement_degree = 1

    problem = dmech.HyperelasticityProblem(
        mesh=mesh,
        define_facet_normals=1,
        boundaries_mf=boundaries_mf,
        points_mf=points_mf,
        displacement_degree=displacement_degree,
        quadrature_degree="default",
        w_incompressibility=incomp,
        elastic_behavior=mat_params)

    ########################################## Boundary conditions & Loading ###

    Deltat = step_params.get("Deltat", 1.)
    dt_ini = step_params.get("dt_ini", 1.)
    dt_min = step_params.get("dt_min", 1.)

    k_step = problem.add_step(
        Deltat=Deltat,
        dt_ini=dt_ini,
        dt_min=dt_min)

    load_type = load_params.get("type", "disp")

    if (load_type == "disp"): # MG20220813: It would be possible to impose the spatially varying displacement directly through an expression, but this would need to be implemented within Constraint, e.g. with a TimeVaryingExpression.
        surface_nodes_coords = [node_coords for node_coords in mesh.coordinates() if dolfin.near((node_coords[0]-X0)**2 + (node_coords[1]-Y0)**2, R**2, eps=1e-3)]
        dR = load_params.get("dR", +0.1)
        for X in surface_nodes_coords:
            X_inplane = numpy.array(X) - numpy.array([X0,Y0])
            R = numpy.linalg.norm(X_inplane)
            T = math.atan2(X_inplane[1], X_inplane[0])
            r = R + dR
            t = T
            x_inplane = numpy.array([r * math.cos(t), r * math.sin(t)])
            x = numpy.array([X0,Y0]) + x_inplane
            U = x - X
            # X_sd = dolfin.AutoSubDomain(lambda x, on_boundary: dolfin.near(x[0], X[0], eps=1e-3) and dolfin.near(x[1], X[1], eps=1e-3)) # MG20220813: OMG this behaves so weird!
            X_sd = dolfin.CompiledSubDomain("near(x[0], x0) && near(x[1], y0)", x0=X[0], y0=X[1])
            problem.add_constraint(
                V=problem.get_displacement_function_space(),
                sub_domain=X_sd,
                val_ini=[0.,0.], val_fin=U,
                k_step=k_step,
                method="pointwise")
    elif (load_type == "pres"):
        problem.add_constraint(
            V=problem.get_displacement_function_space().sub(1),
            sub_domain=x1_sd,
            val=0.,
            method="pointwise")
        problem.add_constraint(
            V=problem.get_displacement_function_space().sub(0),
            sub_domain=x2_sd,
            val=0.,
            method="pointwise")
        problem.add_constraint(
            V=problem.get_displacement_function_space().sub(1),
            sub_domain=x3_sd,
            val=0.,
            method="pointwise")
        problem.add_constraint(
            V=problem.get_displacement_function_space().sub(0),
            sub_domain=x4_sd,
            val=0.,
            method="pointwise")
        p = load_params.get("p", -1.0)
        problem.add_surface_pressure_loading_operator(
            measure=problem.dS(S_id),
            P_ini=0, P_fin=p,
            k_step=k_step)

    ################################################# Quantities of Interest ###

    problem.add_point_displacement_qoi(
        name="U",
        coordinates=[X0+R, Y0],
        component=0)

    ################################################################# Solver ###

    solver = dmech.NonlinearSolver(
        problem=problem,
        parameters={
            "sol_tol":[1e-6]*len(problem.subsols),
            "n_iter_max":32},
        relax_type="constant",
        write_iter=0)

    integrator = dmech.TimeIntegrator(
        problem=problem,
        solver=solver,
        parameters={
            "n_iter_for_accel":4,
            "n_iter_for_decel":16,
            "accel_coeff":2,
            "decel_coeff":2},
        print_out=res_basename*verbose,
        print_sta=res_basename*verbose,
        write_qois=res_basename+"-qois",
        write_qois_limited_precision=1,
        write_sol=res_basename*verbose,
        write_vtus=res_basename*verbose,
        write_vtus_with_preserved_connectivity=write_vtus_with_preserved_connectivity)

    success = integrator.integrate()
    assert (success),\
        "Integration failed. Aborting."

    integrator.close()



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class SurfaceTensionLoadingOperator(Operator):

    def __init__(self,
            U,
            U_test,
            kinematics,
            N,
            measure,
            gamma_val=None, gamma_ini=None, gamma_fin=None):

        self.measure = measure

        self.tv_gamma = dmech.TimeVaryingConstant(
            val=gamma_val, val_ini=gamma_ini, val_fin=gamma_fin)
        gamma = self.tv_gamma.val

        FmTN = dolfin.dot(dolfin.inv(kinematics.F).T, N)
        T = dolfin.sqrt(dolfin.inner(FmTN, FmTN))
        Pi = gamma * T * kinematics.J * self.measure
        self.res_form = dolfin.derivative(Pi, U, U_test)



    def set_value_at_t_step(self,
            t_step):

        self.tv_gamma.set_value_at_t_step(t_step)

################################################################################

class SurfaceTension0LoadingOperator(Operator):

    def __init__(self,
            u,
            u_test,
            kinematics,
            N,
            measure,
            gamma_val=None, gamma_ini=None, gamma_fin=None):

        self.measure = measure

        self.tv_gamma = dmech.TimeVaryingConstant(
            val=gamma_val, val_ini=gamma_ini, val_fin=gamma_fin)
        gamma = self.tv_gamma.val

        dim = u.ufl_shape[0]
        I = dolfin.Identity(dim)
        Pi = gamma * (1 + dolfin.inner(
            kinematics.epsilon,
            I - dolfin.outer(N,N))) * self.measure
        self.res_form = dolfin.derivative(Pi, u, u_test) # MG20211220: Is that correct?!



    def set_value_at_t_step(self,
            t_step):

        self.tv_gamma.set_value_at_t_step(t_step)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import numpy

import dolfin_mech as dmech

################################################################################

class Kinematics():



    def __init__(self,
            U,
            U_old=None,
            Q_expr=None):

        self.U = U

        self.dim = self.U.ufl_shape[0]
        self.I = dolfin.Identity(self.dim)

        self.F     = self.I + dolfin.grad(self.U)
        self.F     = dolfin.variable(self.F)
        self.J     = dolfin.det(self.F)
        self.C     = self.F.T * self.F
        self.C     = dolfin.variable(self.C)
        self.C_inv = dolfin.inv(self.C)
        self.IC    = dolfin.tr(self.C)
        self.IIC   = (dolfin.tr(self.C)*dolfin.tr(self.C) - dolfin.tr(self.C*self.C))/2
        self.E     = (self.C - self.I)/2
        self.E     = dolfin.variable(self.E)

        self.E_sph = dolfin.tr(self.E)/self.dim * self.I
        self.E_dev = self.E - self.E_sph

        self.F_bar     = self.J**(-1/self.dim) * self.F
        self.C_bar     = self.F_bar.T * self.F_bar
        self.C_bar_inv = dolfin.inv(self.C_bar)
        self.IC_bar    = dolfin.tr(self.C_bar)
        self.IIC_bar   = (dolfin.tr(self.C_bar)*dolfin.tr(self.C_bar) - dolfin.tr(self.C_bar*self.C_bar))/2
        self.E_bar     = (self.C_bar - self.I)/2

        if (U_old is not None):
            self.U_old = U_old

            self.F_old = self.I + dolfin.grad(U_old)
            self.J_old = dolfin.det(self.F_old)
            self.C_old = self.F_old.T * self.F_old
            self.E_old = (self.C_old - self.I)/2

            self.F_bar_old = self.J_old**(-1/self.dim) * self.F_old
            self.C_bar_old = self.F_bar_old.T * self.F_bar_old
            self.E_bar_old = (self.C_bar_old - self.I)/2

            self.F_mid = (self.F_old + self.F)/2
            self.J_mid = (self.J_old + self.J)/2
            self.C_mid = (self.C_old + self.C)/2
            self.E_mid = (self.E_old + self.E)/2

            self.F_bar_mid = (self.F_bar_old + self.F_bar)/2
            self.C_bar_mid = (self.C_bar_old + self.C_bar)/2
            self.E_bar_mid = (self.E_bar_old + self.E_bar)/2

        if (Q_expr is not None):
            self.Q_expr = Q_expr

            self.E_loc     = self.Q_expr * self.E     * self.Q_expr.T # MG20211215: This should work, right?
            self.E_bar_loc = self.Q_expr * self.E_bar * self.Q_expr.T # MG20211215: This should work, right?
            # self.E_loc     = dolfin.dot(dolfin.dot(self.Q_expr, self.E    ), self.Q_expr.T)
            # self.E_bar_loc = dolfin.dot(dolfin.dot(self.Q_expr, self.E_bar), self.Q_expr.T)

            if (U_old is not None):
                self.E_old_loc     = self.Q_expr * self.E_old     * self.Q_expr.T # MG20211215: This should work, right?
                self.E_bar_old_loc = self.Q_expr * self.E_bar_old * self.Q_expr.T # MG20211215: This should work, right?
                # self.E_old_loc     = dolfin.dot(dolfin.dot(self.Q_expr, self.E_old    ), self.Q_expr.T)
                # self.E_bar_old_loc = dolfin.dot(dolfin.dot(self.Q_expr, self.E_bar_old), self.Q_expr.T)



import datetime
import os
import setuptools

# version = os.environ['CI_COMMIT_TAG']
version = datetime.date.today().strftime("%Y.%m.%d")

setuptools.setup(
    name="dolfin_mech",
    version=version,
    author="Martin Genet",
    author_email="martin.genet@polytechnique.edu",
    description=open("README.md", "r").readlines()[1][:-1],
    long_description = open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://gitlab.inria.fr/mgenet/dolfin_mech",
    packages=["dolfin_mech"],
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=["gmsh", "matplotlib", "meshio", "myPythonLibrary", "myVTKPythonLibrary", "numpy", "pandas", "vtk"],
)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import numpy

import dolfin_mech as dmech

################################################################################

class LinearizedKinematics():



    def __init__(self,
            u,
            u_old=None):

        self.u = u

        self.dim = self.u.ufl_shape[0]
        self.I = dolfin.Identity(self.dim)

        self.epsilon = dolfin.sym(dolfin.grad(self.u))
        self.epsilon = dolfin.variable(self.epsilon)

        self.epsilon_sph = dolfin.tr(self.epsilon)/self.dim * self.I
        self.epsilon_dev = self.epsilon - self.epsilon_sph

        if (u_old is not None):
            self.u_old = u_old
            
            self.epsilon_old = dolfin.sym(dolfin.grad(self.u_old))

            self.epsilon_sph_old = dolfin.tr(self.epsilon_old)/self.dim * self.I
            self.epsilon_dev_old = self.epsilon_old - self.epsilon_sph_old

            self.epsilon_mid = (self.epsilon_old + self.epsilon)/2



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Cécile Patte, 2019-2021                                              ###
###                                                                          ###
### INRIA, Palaiseau, France                                                 ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class WbulkLungElasticMaterial(ElasticMaterial):



    def __init__(self,
            Phis,
            Phis0,
            parameters):

        assert ('kappa' in parameters)
        self.kappa = dolfin.Constant(parameters['kappa'])

        Phis = dolfin.variable(Phis)
        self.Psi = self.kappa * (Phis/Phis0 - 1 - dolfin.ln(Phis/Phis0))
        self.dWbulkdPhis = dolfin.diff(self.Psi, Phis)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class WskelLungElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters):

        self.kinematics = kinematics

        self.bulk = dmech.ExponentialOgdenCiarletGeymonatElasticMaterial(kinematics, parameters)
        self.dev  = dmech.NeoHookeanMooneyRivlinElasticMaterial(kinematics, parameters)

        self.Psi   = self.bulk.Psi   + self.dev.Psi
        self.Sigma = self.bulk.Sigma + self.dev.Sigma
        self.P     = self.bulk.P     + self.dev.P
        self.sigma = self.bulk.sigma + self.dev.sigma



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class LinearizedElasticityOperator(Operator):

    def __init__(self,
            kinematics,
            u_test,
            material_model,
            material_parameters,
            measure):

        self.kinematics = kinematics
        self.material   = dmech.material_factory(kinematics, material_model, material_parameters)
        self.measure    = measure

        epsilon_test = dolfin.sym(dolfin.grad(u_test))
        self.res_form = dolfin.inner(self.material.sigma, epsilon_test) * self.measure



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class OgdenCiarletGeymonatElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters,
            decoup=False):

        self.kinematics = kinematics

        self.C0 = self.get_C0_from_parameters(parameters, decoup)

        self.Psi = self.C0 * (self.kinematics.J**2 - 1 - 2*dolfin.ln(self.kinematics.J)) # MG20180516: In 2d, plane strain

        self.checkJ = parameters.get("checkJ", False)
        if (self.checkJ):
            self.Sigma = dolfin.conditional( # MG20230320: Otherwise Sigma is well defined for J < 0…
                dolfin.gt(self.kinematics.J, 0.),
                2*self.C0 * (self.kinematics.J**2 - 1) * self.kinematics.C_inv, # MG20200206: Cannot differentiate Psi wrt to C because J is not defined as a function of C
                self.kinematics.C_inv/dolfin.Constant(0.))
            
        else:
            self.Sigma = 2*self.C0 * (self.kinematics.J**2 - 1) * self.kinematics.C_inv # MG20200206: Cannot differentiate Psi wrt to C because J is not defined as a function of C

        if (self.kinematics.dim == 2):
            self.Sigma_ZZ = 2*self.C0 * (self.kinematics.J**2 - 1)

        # self.P = dolfin.diff(self.Psi, self.kinematics.F) # MG20220426: Cannot do that for micromechanics problems
        # self.P = 2*self.C0 * (self.kinematics.J**2 - 1) * self.kinematics.F_inv.T
        self.P = self.kinematics.F * self.Sigma

        self.sigma = self.P * self.kinematics.F.T / self.kinematics.J
        


    # def get_free_energy(self,
    #         U=None,
    #         C=None):

    #     C  = self.get_C_from_U_or_C(U, C)
    #     JF = dolfin.sqrt(dolfin.det(C)) # MG20200207: Watch out! This is well defined for inverted elements!

    #     Psi   = self.C0 * (JF**2 - 1 - 2*dolfin.ln(JF)) # MG20180516: in 2d, plane strain
    #     Sigma = 2*dolfin.diff(Psi, C)

    #     # C_inv = dolfin.inv(C)
    #     # Sigma = 2*self.C0 * (JF**2 - 1) * C_inv

    #     return Psi, Sigma



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Mahdi Manoochehrtayebi, 2020-2024                                    ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import numpy

import dolfin_mech as dmech

################################################################################

def run_HollowBox_MicroPoroHyperelasticity(
        dim,
        mesh=None,
        mesh_params=None,
        displacement_perturbation_degree=1,
        quadrature_degree=3,
        mat_params={},
        bcs="pbc",
        step_params={},
        load_params={},
        res_basename="run_HollowBox_MicroPoroHyperelasticity",
        add_p_hydro_and_Sigma_VM_FoI=False,
        write_qois_limited_precision=True,
        verbose=0):

    assert ((mesh is not None) ^ (mesh_params is not None))
    if (mesh is None):
        mesh = dmech.run_HollowBox_Mesh(
            params=mesh_params)

    coord = mesh.coordinates()
    xmax = max(coord[:,0]); xmin = min(coord[:,0])
    ymax = max(coord[:,1]); ymin = min(coord[:,1])
    if (dim==2):    
        bbox = [xmin, xmax, ymin, ymax]
        vertices = numpy.array([[xmin, ymin],
                                [xmax, ymin],
                                [xmax, ymax],
                                [xmin, ymax]])
        a1 = vertices[1,:]-vertices[0,:] # first vector generating periodicity
        a2 = vertices[3,:]-vertices[0,:] # second vector generating periodicity
        # check if UC vertices form indeed a parallelogram
        tol = 1E-8
        assert numpy.linalg.norm(vertices[2, :]-vertices[3, :] - a1) <= tol
        assert numpy.linalg.norm(vertices[2, :]-vertices[1, :] - a2) <= tol
    elif (dim==3):    
        zmax = max(coord[:,2]); zmin = min(coord[:,2])
        bbox = [xmin, xmax, ymin, ymax, zmin, zmax]
        vertices = numpy.array([[xmin, ymin, zmin],
                                [xmax, ymin, zmin],
                                [xmax, ymax, zmin],
                                [xmin, ymax, zmin],
                                [xmin, ymin, zmax],
                                [xmax, ymin, zmax],
                                [xmax, ymax, zmax],
                                [xmin, ymax, zmax]])

    ################################################## Subdomains & Measures ###

    tol = 1E-8
    xmin_sd = dolfin.CompiledSubDomain("near(x[0], x0, tol) && on_boundary", x0=xmin, tol=tol)
    xmax_sd = dolfin.CompiledSubDomain("near(x[0], x0, tol) && on_boundary", x0=xmax, tol=tol)
    ymin_sd = dolfin.CompiledSubDomain("near(x[1], x0, tol) && on_boundary", x0=ymin, tol=tol)
    ymax_sd = dolfin.CompiledSubDomain("near(x[1], x0, tol) && on_boundary", x0=ymax, tol=tol)
    if (dim==3): zmin_sd = dolfin.CompiledSubDomain("near(x[2], x0) && on_boundary", x0=zmin, tol=tol)
    if (dim==3): zmax_sd = dolfin.CompiledSubDomain("near(x[2], x0) && on_boundary", x0=zmax, tol=tol)

    # if (dim==2):
    #     sint_sd = dolfin.CompiledSubDomain("near(pow(x[0] - x0, 2) + pow(x[1] - y0, 2), pow(r0, 2), 1e-2) && on_boundary", x0=x0, y0=y0, r0=r0)
    # elif (dim==3):
    #     sint_sd = dolfin.CompiledSubDomain("near(pow(x[0] - x0, 2) + pow(x[1] - y0, 2) + pow(x[2] - z0, 2), pow(r0, 2), 1e-2) && on_boundary", x0=x0, y0=y0, z0=z0, r0=r0)

    xmin_id = 1
    xmax_id = 2
    ymin_id = 3
    ymax_id = 4
    if (dim==3): zmin_id = 5
    if (dim==3): zmax_id = 6
    # sint_id = 9

    boundaries_mf = dolfin.MeshFunction("size_t", mesh, mesh.topology().dim()-1) # MG20180418: size_t looks like unisgned int, but more robust wrt architecture and os
    boundaries_mf.set_all(0)

    xmin_sd.mark(boundaries_mf, xmin_id)
    xmax_sd.mark(boundaries_mf, xmax_id)
    ymin_sd.mark(boundaries_mf, ymin_id)
    ymax_sd.mark(boundaries_mf, ymax_id)
    if (dim==3): zmin_sd.mark(boundaries_mf, zmin_id)
    if (dim==3): zmax_sd.mark(boundaries_mf, zmax_id)
    # sint_sd.mark(boundaries_mf, sint_id)

    if (verbose):
        xdmf_file_boundaries = dolfin.XDMFFile(res_basename+"-boundaries.xdmf")
        xdmf_file_boundaries.write(boundaries_mf)
        xdmf_file_boundaries.close()

    ################################################################ Problem ###

    problem = dmech.MicroPoroHyperelasticityProblem(
        mesh=mesh,
        mesh_bbox=bbox,
        vertices=vertices,
        boundaries_mf=boundaries_mf,
        displacement_perturbation_degree=displacement_perturbation_degree,
        quadrature_degree=quadrature_degree,
        solid_behavior=mat_params,
        bcs=bcs)

    ################################################################ Loading ###

    n_steps = step_params.get("n_steps", 1)
    Deltat_lst = step_params.get("Deltat_lst", [step_params.get("Deltat", 1.)/n_steps]*n_steps)
    dt_ini_lst = step_params.get("dt_ini_lst", [step_params.get("dt_ini", 1.)/n_steps]*n_steps)
    dt_min_lst = step_params.get("dt_min_lst", [step_params.get("dt_min", 1.)/n_steps]*n_steps)
    dt_max_lst = step_params.get("dt_max_lst", [step_params.get("dt_max", 1.)/n_steps]*n_steps)

    U_bar_ij_lst = [[None for i in range(dim)] for j in range(dim)]
    sigma_bar_ij_lst = [[None for i in range(dim)] for j in range(dim)]
    for i in range(dim):
     for j in range (dim):
        U_bar_ij_lst[i][j] = load_params.get("U_bar_"+str(i)+str(j)+"_lst", [load_params.get("U_bar_"+str(i)+str(j), None) for k_step in range(n_steps)])
        sigma_bar_ij_lst[i][j] = load_params.get("sigma_bar_"+str(i)+str(j)+"_lst", [load_params.get("sigma_bar_"+str(i)+str(j), None) for k_step in range(n_steps)])
    pf_lst = load_params.get("pf_lst", [(k_step+1)*load_params.get("pf", 0)/n_steps for k_step in range(n_steps)])

    for k_step in range(n_steps):

        Deltat = Deltat_lst[k_step]
        dt_ini = dt_ini_lst[k_step]
        dt_min = dt_min_lst[k_step]
        dt_max = dt_max_lst[k_step]

        k_step = problem.add_step(
            Deltat=Deltat,
            dt_ini=dt_ini,
            dt_min=dt_min,
            dt_max=dt_max)

        pf = pf_lst[k_step]
        pf_old = pf_lst[k_step-1] if (k_step > 0) else 0.
        problem.add_surface_pressure_loading_operator(
            measure=problem.dS(0),
            P_ini=pf_old, P_fin=pf,
            k_step=k_step)

        for i in range(dim):
         for j in range (dim):
            U_bar_ij = U_bar_ij_lst[i][j][k_step]
            U_bar_ij_old = U_bar_ij_lst[i][j][k_step-1] if (k_step > 0) else 0.
            sigma_bar_ij = sigma_bar_ij_lst[i][j][k_step]
            sigma_bar_ij_old = sigma_bar_ij_lst[i][j][k_step-1] if (k_step > 0) else 0.
            assert ((U_bar_ij is not None) ^ (sigma_bar_ij is not None))
            if (U_bar_ij is not None):
                problem.add_macroscopic_stretch_component_penalty_operator(
                    i=i, j=j,
                    U_bar_ij_ini=U_bar_ij_old, U_bar_ij_fin=U_bar_ij,
                    pen_val=1e6,
                    k_step=k_step)
            elif (sigma_bar_ij is not None):
                problem.add_macroscopic_stress_component_constraint_operator(
                    i=i, j=j,
                    sigma_bar_ij_ini=sigma_bar_ij_old, sigma_bar_ij_fin=sigma_bar_ij,
                    pf_ini=pf_old, pf_fin=pf,
                    k_step=k_step)

    ################################################# Quantities of Interest ###

    problem.add_deformed_solid_volume_qoi()
    problem.add_deformed_fluid_volume_qoi()
    problem.add_deformed_volume_qoi()
    problem.add_macroscopic_stretch_qois()
    problem.add_macroscopic_solid_stress_qois()
    problem.add_macroscopic_solid_hydrostatic_pressure_qoi()
    problem.add_macroscopic_stress_qois()
    problem.add_fluid_pressure_qoi()

    if (add_p_hydro_and_Sigma_VM_FoI):
        for operator in problem.operators:
            if hasattr(operator, "material"):
                material = operator.material
                break
        problem.add_foi(expr=material.p_hydro, fs=problem.sfoi_fs, name="p_hydro")
        problem.add_foi(expr=material.Sigma_VM, fs=problem.sfoi_fs, name="Sigma_VM")

    ################################################################# Solver ###

    solver = dmech.NonlinearSolver(
        problem=problem,
        parameters={
            "sol_tol":[1e-6]*len(problem.subsols),
            "n_iter_max":32},
        relax_type="constant",
        write_iter=0)

    integrator = dmech.TimeIntegrator(
        problem=problem,
        solver=solver,
        parameters={
            "n_iter_for_accel":4,
            "n_iter_for_decel":16,
            "accel_coeff":2,
            "decel_coeff":2},
        print_out=res_basename*verbose,
        print_sta=res_basename*verbose,
        write_qois=res_basename+"-qois",
        write_qois_limited_precision=write_qois_limited_precision,
        write_sol=res_basename*verbose)

    success = integrator.integrate()
    assert (success),\
        "Integration failed. Aborting."

    integrator.close()

    return problem
